
gaze_data_fname                 = 'gazeData.tsv'
frame_timestamps_fname          = 'frameTimestamps.tsv'
scene_camera_video_fname_stem   = 'worldCamera'
scene_camera_calibration_fname  = 'calibration.xml'